@@ -16,6 +16,14 @@ import pandas as pd
 from multiprocessing.pool import ThreadPool
 import nibabel as nib
 import sys
+from tensorboardX import SummaryWriter
+from torch.utils.data import Dataset
+from torch.utils.data import DataLoader
+import torch
+import torch.utils.model_zoo as model_zoo
+from torchvision.models.resnet import BasicBlock
+from torch import nn
+import math
 
 __author__ = "Junhao Wen"
 __copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
@@ -1066,4 +1074,306 @@ def time_bar(i,num):
     progress = (i+1)*1.0/num
     prog_int = int(progress*50)
     sys.stdout.write('\t\t[%s%s] %.2f%%' % ('='*prog_int,' '*(50-prog_int), progress*100))
-    sys.stdout.flush()
\ No newline at end of file
+    sys.stdout.flush()
+
+###### Neural Network for Regression, e.g., age prediction
+class neural_network_regression(nn.Module):
+    def __init__(self, input_dim):
+        super(neural_network_regression, self).__init__()
+        self.input_layer = nn.Linear(input_dim, 128)
+        self.hidden_layer1 = nn.Linear(128, 64)
+        self.output_layer = nn.Linear(64, 1)
+        self.relu = nn.ReLU()
+
+    def forward(self, x):
+        out = self.relu(self.input_layer(x))
+        out = self.relu(self.hidden_layer1(out))
+        out = self.output_layer(out)
+        return out
+
+
+##### Resnet adapted to regerssion
+model_urls = {
+    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth'
+}
+
+class DatasetShuffler(Dataset):
+    """
+    This class reads the batch size of samples
+
+    Return: a Pytorch Dataset objective
+    """
+
+    def __init__(self, data_tensor_x, data_tensor_y, df_header, transformations=False):
+        """
+        Args:
+            data_tensor (string): data tensor.
+            transformations (callable, optional): if the data sample should be done some transformations or not, such as resize the image.
+        """
+        self.data_tensor_x = data_tensor_x
+        self.data_tensor_y = data_tensor_y
+        self.df_header = df_header
+        self.transformations = transformations
+        self.df_x = data_tensor_x
+        self.df_y = data_tensor_y
+
+    def __len__(self):
+        return self.df_x.shape[0]
+
+    def __getitem__(self, idx):
+        x = self.data_tensor_x[idx, :]
+        y = self.data_tensor_y[idx]
+        participant_id = self.df_header.iloc[idx, :][0]
+        session_id = self.df_header.iloc[idx, :][1]
+
+        if self.transformations:
+            x = self.transformations(x)
+
+        sample = {'x': x, 'y': y, 'participant_id': participant_id, 'session_id': session_id}
+
+        return sample
+
+def train(model, data_loader, use_cuda, loss_func, optimizer, writer, epoch, batch_size, model_mode="train"):
+    """
+    This is the function to train, validate or test the model, depending on the model_mode parameter.
+    :param model:
+    :param data_loader:
+    :param use_cuda:
+    :param loss_func:
+    :param optimizer:
+    :param writer:
+    :param epoch:
+    :return:
+    """
+    global_step = None
+
+    if model_mode == "train":
+        columns = ['participant_id', 'session_id', 'true_label', 'predicted_label']
+        results_df = pd.DataFrame(columns=columns)
+        total_loss = 0.0
+
+        model.train()  # set the model to training mode
+        print('The number of batches in this sampler based on the batch size: %s' % str(len(data_loader)))
+
+        for i, data in enumerate(data_loader):
+            # update the global step
+            global_step = i + epoch * len(data_loader)
+
+            if use_cuda:
+                imgs, labels = data['x'].cuda(), data['y'].cuda()
+            else:
+                imgs, labels = data['x'], data['y']
+
+
+            predicted = torch.squeeze(model(imgs))
+            batch_loss = loss_func(predicted, labels)
+            total_loss += batch_loss.item()
+            writer.add_scalar('loss', batch_loss.item()/batch_size, global_step)
+
+            optimizer.zero_grad()
+            batch_loss.backward()
+            optimizer.step()
+
+            # Generate detailed DataFrame
+            for idx, sub in enumerate(data['participant_id']):
+                row = [sub, data['session_id'][idx],
+                       labels[idx].item(), predicted[idx].item()]
+                row_df = pd.DataFrame(np.array(row).reshape(1, -1), columns=columns)
+                results_df = pd.concat([results_df, row_df])
+
+            # delete the temporary variables taking the GPU memory
+            del imgs, labels, predicted, batch_loss
+            torch.cuda.empty_cache()
+
+        loss_batch_mean = total_loss / len(data_loader)
+        torch.cuda.empty_cache()
+
+    elif model_mode == "valid":
+        results_df, total_loss = test(model, data_loader, use_cuda, loss_func)
+        loss_batch_mean = total_loss / len(data_loader)
+        writer.add_scalar('loss', loss_batch_mean/batch_size, epoch)
+        torch.cuda.empty_cache()
+
+    else:
+        raise ValueError('This mode %s was not implemented. Please choose between train and valid' % model_mode)
+
+    return results_df, loss_batch_mean/batch_size, global_step
+
+def test(model, data_loader, use_cuda, loss_func):
+    """
+    The function to evaluate the testing data for the trained classifiers
+    :param model:
+    :param data_loader:
+    :param use_cuda:
+    :return:
+    """
+
+    columns = ['participant_id', 'session_id', 'true_label', 'predicted_label']
+    results_df = pd.DataFrame(columns=columns)
+    total_loss = 0
+
+    if use_cuda:
+        model.cuda()
+
+    model.eval()  # set the model to evaluation mode
+    torch.cuda.empty_cache()
+    with torch.no_grad():
+        for i, data in enumerate(data_loader):
+            if use_cuda:
+                imgs, labels = data['x'].cuda(), data['y'].cuda()
+            else:
+                imgs, labels = data['x'], data['y']
+
+            predicted = torch.squeeze(model(imgs))
+            loss = loss_func(predicted, labels)
+            total_loss += loss.item()
+
+            # Generate detailed DataFrame
+            for idx, sub in enumerate(data['participant_id']):
+                row = [sub, data['session_id'][idx],
+                       labels[idx].item(), predicted[idx].item()]
+
+                row_df = pd.DataFrame(np.array(row).reshape(1, -1), columns=columns)
+                results_df = pd.concat([results_df, row_df])
+
+            del imgs, labels
+            torch.cuda.empty_cache()
+
+        results_df.reset_index(inplace=True, drop=True)
+        torch.cuda.empty_cache()
+
+    return results_df, total_loss
+
+def save_checkpoint(state, loss_is_best, checkpoint_dir, filename='checkpoint.pth.tar'):
+    import torch
+    import os
+
+    if not os.path.exists(checkpoint_dir):
+        os.makedirs(checkpoint_dir)
+    if loss_is_best:
+        torch.save(state, os.path.join(checkpoint_dir, filename))
+
+def load_model(model, checkpoint_dir, gpu, filename='model_best.pth.tar'):
+    """
+    Load the weights written in checkpoint_dir in the model object.
+
+    :param model: (Module) CNN in which the weights will be loaded.
+    :param checkpoint_dir: (str) path to the folder containing the parameters to loaded.
+    :param gpu: (bool) if True a gpu is used.
+    :param filename: (str) Name of the file containing the parameters to loaded.
+    :return: (Module) the update model.
+    """
+    from copy import deepcopy
+    import torch
+    import os
+
+    best_model = deepcopy(model)
+    param_dict = torch.load(os.path.join(checkpoint_dir, filename), map_location="cpu")
+    best_model.load_state_dict(param_dict['model'])
+
+    if gpu:
+        best_model = best_model.cuda()
+
+    return best_model, param_dict['epoch']
+
+def to_tsvs(output_dir, results_df, result, fold, dataset='train'):
+    """
+    Allows to save the outputs of the test function.
+
+    :param output_dir: (str) path to the output directory.
+    :param results_df: (DataFrame) the individual results per slice.
+    :param result: (float) MSE per batch.
+    :param fold: (int) the fold for which the performances were obtained.
+    :param selection: (str) the metrics on which the model was selected (best_acc, best_loss)
+    :param dataset: (str) the dataset on which the evaluation was performed.
+    :return:
+    """
+    performance_dir = os.path.join(output_dir, 'performances', 'fold_%i' % fold)
+
+    if not os.path.exists(performance_dir):
+        os.makedirs(performance_dir)
+
+    results_df.to_csv(os.path.join(performance_dir, dataset + '.tsv'), index=False, sep='\t')
+    pd.DataFrame([result], index=[0]).to_csv(os.path.join(performance_dir, dataset + '_mse.tsv'), index=False, sep='\t')
+
+def train_network(model, output_dir, fi, X_train, y_train, X_test, y_test, num_epochs,
+                  batch_size, init_state, df_header, gpu, lr, weight_decay, opt):
+
+    print("Running for the %d-th repetition of the repeated holdout CV" % fi)
+    writer_train_batch = SummaryWriter(log_dir=(os.path.join(output_dir, "log_dir", "fold_%i" % fi,
+                                                             "train_batch")))
+    writer_train_all_data = SummaryWriter(log_dir=(os.path.join(output_dir, "log_dir", "fold_%i" % fi,
+                                                                "train_all_data")))
+    writer_valid = SummaryWriter(log_dir=(os.path.join(output_dir, "log_dir", "fold_%i" % fi, "valid")))
+
+    data_train = DatasetShuffler(X_train, y_train, df_header)
+    data_valid = DatasetShuffler(X_test, y_test, df_header)
+
+    # Use argument load to distinguish training and testing
+    train_loader = DataLoader(data_train,
+                              batch_size=batch_size,
+                              shuffle=True,
+                              num_workers=4,
+                              pin_memory=True)
+
+    valid_loader = DataLoader(data_valid,
+                              batch_size=batch_size,
+                              shuffle=False,
+                              num_workers=4,
+                              pin_memory=True)
+
+    # Define Optimizer and Loss Function
+    optimizer = eval("torch.optim." + opt)(filter(lambda x: x.requires_grad, model.parameters()),
+                                                        lr=lr, weight_decay=weight_decay)
+    loss = torch.nn.MSELoss()
+
+    model.load_state_dict(init_state)
+
+    # parameters used in training
+    best_loss_valid = np.inf
+
+    for epoch in range(num_epochs):
+        print("At %i-th epoch." % epoch)
+
+        # train the model
+        train_df, mean_mse_train, global_step = train(model, train_loader, gpu, loss, optimizer, writer_train_batch, epoch, batch_size, model_mode='train')
+        print("For training, mean MSE loss: %f during each epoch using batch data %d" % (mean_mse_train, epoch))
+
+        # calculate the accuracy with the whole training data for subject level balanced accuracy
+        train_all_df, mean_mse_train_all, _ = train(model, train_loader, gpu, loss, optimizer, writer_train_all_data, epoch, batch_size, model_mode='valid')
+        print("For training, mean MSE loss: %f at the end of epoch by applying all training data %d" % (mean_mse_train_all, epoch))
+
+        # at then end of each epoch, we validate one time for the model with the validation data
+        valid_df, mean_mse_valid, _ = train(model, valid_loader, gpu, loss, optimizer, writer_valid, epoch, batch_size, model_mode='valid')
+        print("For validation, mean MSE loss: %f at the end of epoch %d" % (mean_mse_valid, epoch))
+
+        # save the best model based on the best loss
+        loss_is_best = mean_mse_valid < best_loss_valid
+        best_loss_valid = min(mean_mse_valid, best_loss_valid)
+
+        save_checkpoint({
+            'epoch': epoch + 1,
+            'model': model.state_dict(),
+            'loss': mean_mse_valid,
+            'optimizer': optimizer.state_dict(),
+            'global_step': global_step},
+            loss_is_best,
+            os.path.join(output_dir, "best_model_dir", "fold_" + str(fi), "NN", "best_loss"),
+            filename="model_best.pth.tar")
+
+    # Final evaluation for all criteria
+    model, best_epoch = load_model(model, os.path.join(output_dir, 'best_model_dir', 'fold_%i' % fi, 'NN', 'best_loss'),
+                                   gpu=gpu, filename='model_best.pth.tar')
+
+    train_df, mse_train = test(model, train_loader, gpu, loss)
+    valid_df, mse_valid = test(model, valid_loader, gpu, loss)
+
+    # write the information of subjects and performances into tsv files.
+    to_tsvs(output_dir, train_df, mse_train/batch_size/len(train_loader), fi, dataset='train')
+    to_tsvs(output_dir, valid_df, mse_valid/batch_size/len(valid_loader), fi, dataset='validation')
+
+    del optimizer, writer_train_batch, writer_train_all_data, writer_valid
+    torch.cuda.empty_cache()
+    global_step = 0
+
+
